#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@Time    :   2023/09/11 10:19:27
@Author  :   ChenHao
@Contact :   jerrychen1990@gmail.com
'''


import copy
import itertools
import os
from typing import Any, List, Union
from cachetools import LRUCache, cached
from loguru import logger

import numpy as np
import zhipuai
from agit import AGIT_LOG_HOME
from snippets import retry, jdumps
from zhipuai import ZhipuAI


from agit.utils import gen_req_id

fmt = "<green>{time:YYYY-MM-DD HH:mm:ss.SSS}</green> <level>{level: <8}</level>|  - <level>[{extra[request_id]}]{message}</level>"

logger.add(os.path.join(AGIT_LOG_HOME, "zhipuai_bk.log"), retention="365 days", rotation=" 00:00", level="DEBUG", format=fmt,
           filter=lambda r: r["function"] == "call_llm_api_v4")

default_logger = logger


def check_api_key(api_key):
    if api_key is None:
        api_key = os.environ.get("ZHIPUAI_API_KEY", None)
    if not api_key:
        raise ValueError("未设置api_key且未设置ZHIPUAI_API_KEY环境变量")
    zhipuai.api_key = api_key


def resp_generator(events, max_len=None, err_resp=None):
    token_left = max_len if max_len else None
    for event in events:
        if event.event == "add":
            if token_left:
                if len(event.data) > token_left:
                    data = event.data[:token_left]
                else:
                    data = event.data
                token_left -= len(data)
                yield data
                if token_left == 0:
                    break
            else:
                yield event.data
        elif event.event == "finish":
            pass
        else:
            default_logger.error(
                f"zhipu api resp failed with event:{event.event}, data:{event.data}")
            yield err_resp if err_resp else event.data


def support_system(model: str):
    return model.startswith("chatglm3") or model in ["glm-4", "glm-3-turbo", "glm-4v"]


def resp2generator_v4(resp, logger, request_id):
    tool_calls = None
    acc_chunks = []

    for chunk in resp:
        choices = chunk.choices
        if choices:
            choice = choices[0]
            tool_calls = choice.delta.tool_calls
            acc_chunks.append(chunk)
        break
    # print(acc_chunks)

    def gen():
        acc = []
        for chunk in itertools.chain(acc_chunks, resp):
            choices = chunk.choices
            if choices:
                choice = choices[0]
                if choice.delta.content:
                    delta_content = choice.delta.content
                    yield delta_content
                    acc.append(delta_content)
        resp_msg = "".join(acc).strip()
        with logger.contextualize(request_id=request_id):
            logger.debug(f"model generate answer:{resp_msg}")

    return tool_calls, gen()


@cached(LRUCache(maxsize=1000))
def get_client(api_key: str) -> ZhipuAI:
    logger.info(f"new zhipuai client with {api_key=}")
    client = ZhipuAI(api_key=api_key or os.environ.get("ZHIPU_API_KEY"))
    return client


def _is_multimodal(model:str):
    return model in ["glm-4v"]
    
    

def build_messages(prompt: Union[str, dict], model:str, system:str, history:List[dict], role:str, image_url:str, logger):
    if isinstance(prompt, dict):
        messages = history + [prompt]
    else:
        if image_url and _is_multimodal(model):
            message = dict(role=role, content=[dict(type="text", text=prompt), dict(type="image_url", image_url=dict(url=image_url))])
        else:
            message = dict(role=role, content=prompt)
        
        messages = history + [message]   
    if system:
        if support_system(model):
            messages = [dict(role="system", content=system)] + messages
        else:
            logger.warning(f"{model} not support system message, system argument will not work")
    detail_msgs = []
    for idx, item in enumerate(messages):
        detail_msgs.append(f"[{idx+1}].{item['role']}:{item['content']}")
    logger.debug("\n"+"\n".join(detail_msgs))
    
    return messages
    


def call_llm_api(prompt: Union[str, dict], model: str, api_key=None, role="user", image_url:str=None,
                 system=None, history=[], tools=[], do_search=False, search_query=None,
                 logger=None, stream=True, return_origin=False, return_tool_call=False, **kwargs) -> Any:
    import inspect
    if "request_id" not in kwargs:
        request_id = gen_req_id(prompt=prompt, model=model)
        kwargs.update(request_id=request_id)
    request_id = kwargs["request_id"]
    the_logger = logger if logger else default_logger
    
    with the_logger.contextualize(request_id=request_id):
        client = get_client(api_key=api_key or os.environ.get("ZHIPU_API_KEY"))
        valid_keys = dict(inspect.signature(client.chat.completions.create).parameters).keys()
        messages = build_messages(prompt=prompt, model=model, system=system, image_url=image_url, history=history, role=role, logger=the_logger)

        tools = copy.copy(tools)

        if not do_search:
            close_search_tool = {'type': 'web_search', 'web_search': {'enable': False, 'search_query': search_query}}
            tools.append(close_search_tool)
            # the_logger.debug(f"adding close search tool")

        kwargs = {k: v for k, v in kwargs.items() if k in valid_keys}
        # 处理temperature=0的特殊情况
        if kwargs.get("temperature") == 0:
            kwargs.pop("temperature")
            kwargs["do_sample"] = False

        api_inputs = dict(model=model,
                          messages=messages,
                          tools=tools,
                          stream=stream,
                          **kwargs)
        the_logger.debug(f"api_inputs:\n{jdumps(api_inputs)}")

        if return_origin:
            resp = client.chat.completions.create(
                **api_inputs
            )
            the_logger.info(f"api origin resp:{resp}")
            return resp

        api_inputs.update(stream=True)
        response = client.chat.completions.create(
            **api_inputs
        )
        # usage = response.usage
        # the_logger.debug(f"token usage: {usage}")

        tool_call, resp_gen = resp2generator_v4(response, logger=the_logger, request_id=request_id)

        if not stream:
            resp_gen = "".join(resp_gen)
        if return_tool_call:
            return tool_call, resp_gen

        return resp_gen


def call_character_api(prompt: str, user_name, user_info, bot_name, bot_info,
                       history=[], model="characterglm", api_key=None, stream=True, max_len=None, **kwargs):
    check_api_key(api_key)
    zhipu_prompt = history + [dict(role="user", content=prompt)]
    total_words = sum([len(e['content']) for e in zhipu_prompt])
    default_logger.debug(f"zhipu prompt:")
    detail_msgs = []

    for idx, item in enumerate(zhipu_prompt):
        detail_msgs.append(f"[{idx+1}].{item['role']}:{item['content']}")
    default_logger.debug("\n"+"\n".join(detail_msgs))
    default_logger.debug(
        f"{model=},{kwargs=}, history_len={len(history)}, words_num={total_words}")
    meta = {
        "user_info": user_info,
        "bot_info": bot_info,
        "bot_name": bot_name,
        "user_name": user_name
    }
    response = zhipuai.model_api.sse_invoke(
        model=model,
        meta=meta,
        prompt=zhipu_prompt,
        incremental=True,
        ** kwargs
    )
    generator = resp_generator(response.events(), max_len=None)
    if stream:
        return generator
    else:
        resp = "".join(list(generator)).strip()
        if max_len:
            resp = resp[:max_len]
        return resp


def call_embedding_api(text: str, api_key=None, model="text_embedding_v2",
                       norm=None, retry_num=2, wait_time=(1,2)):
    client = get_client(api_key=api_key or os.environ.get("ZHIPU_API_KEY"))

    def zhipu_embedding_attempt():
        resp = client.embeddings.create(
            model=model,  # 填写需要调用的模型名称
            input=text,
        )
        embedding = resp.data[0].embedding
        if norm is not None:
            _norm = 2 if norm == True else norm
            embedding = embedding / np.linalg.norm(embedding, _norm)
        return embedding

    if retry_num:
        attempt = retry(retry_num=retry_num, wait_time=wait_time)(zhipu_embedding_attempt)
    return attempt()


def call_image_gen(prompt: str, api_key: str = None) -> str:
    client = get_client(api_key=api_key or os.environ.get("ZHIPU_API_KEY"))

    response = client.images.generations(
        model="cogview-3",  # 填写需要调用的模型名称
        prompt=prompt,
    )
    return response.data[0].url


if __name__ == "__main__":
    text = "你好"
    system = "你是孔子，请以文言文回答我"
    resp = "".join(call_llm_api(model="chatglm3_130b_int8", prompt=text, system=system, stream=False))
    print(resp)
    emb = call_embedding_api(text)
    print(len(emb))
    print(emb[:4])



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@Time    :   2023/09/11 10:55:05
@Author  :   ChenHao
@Contact :   jerrychen1990@gmail.com
'''
import os
from typing import List

import numpy as np

from agit import AGIT_ENV
from snippets import retry

from loguru import logger as default_logger
from openai import OpenAI




def get_client(proxy:str, api_key:str=None, base_url:str=None):
    if api_key and base_url:
        return OpenAI(api_key=api_key, base_url=base_url)
    if proxy and proxy.lower() == "zhipu":
        default_logger.debug("use zhipu api_key")
        return OpenAI(api_key=os.environ.get("OPENAI_API_KEY_ZHIPU", None), base_url="https://one-api.glm.ai/v1")
    else:
        return OpenAI(api_key=os.environ.get("OPENAI_API_ZHIPU", None))


def get_gen(chunks):
    for chunk in chunks:
        choices = chunk.choices
        if not choices:
            continue
        content = choices[0].delta.content
        yield content or ""


def list_models(keyword=None, api_key=None, base_url=None, proxy=None) -> List[str]:
    client = get_client(api_key=api_key, base_url=base_url, proxy=proxy)
    models = client.models.list().data
    if keyword:
        models = [m for m in models if keyword in m.id]
    return models


def call_llm_api(prompt, model="gpt-3.5-turbo-1106",  history=[],
                 base_url=None, api_key=None, proxy=None, logger=None,
                 system: str = None, tools: List = None,
                 stream=True, **kwargs):

    def _build_messages(prompt, history, system, tools):
        messages = history + [dict(role="user", content=prompt)]
        if system:
            messages = [dict(role="system", content=system)] + messages
        return messages

    logger = logger or default_logger

    messages = _build_messages(prompt, history, system, tools)
    client = get_client(api_key=api_key, base_url=base_url, proxy=proxy)

    detail_msgs = []
    for idx, item in enumerate(messages):
        msg = f"[{idx+1}].{item['role']}:{item['content']}"
        if "tools" in item:
            msg += (f"\n[tools]:{tools}")

        detail_msgs.append(msg)
    logger.debug("\n"+"\n".join(detail_msgs))

    logger.info(f"request to openai with {model=}, {kwargs=}")

    chunks = client.chat.completions.create(
        model=model,
        messages=messages,
        stream=stream,
    )
    gen = get_gen(chunks)
    if stream:
        return gen
    return "".join(gen)


def call_embedding_api(text: str, model="text-embedding-ada-002", api_key=None, base_url=None, proxy=None,
                       norm=None, retry_num=2, wait_time=1):
    client = get_client(api_key=api_key, base_url=base_url, proxy=proxy)

    def attempt():
        embedding = client.embeddings.create(input=[text], model=model).data[0].embedding
        if norm is not None:
            _norm = 2 if norm == True else norm
            embedding = embedding / np.linalg.norm(embedding, _norm)
        return embedding

    if retry_num:
        attempt = retry(retry_num=retry_num, wait_time=wait_time)(attempt)
    return attempt()


if __name__ == "__main__":
    text = "你好,你是谁？"
    model = "gpt-4"
    resp = call_llm_api(text, stream=True, model=model, proxy="ZHIPU")
    for item in resp:
        print(item, end="")

    # resp = call_llm_api(prompt=text, system="把我的话翻译成法语", stream=True, model="gpt-4", proxy="zhipu")
    # for item in resp:
    #     print(item, end="")

    # models = list_models(keyword="embed", proxy="zhipu")
    # # print(models)
    # for model in models:
    #     print(model.id)

    # embd = call_embedding_api(text=text, proxy="zhipu")
    # print(embd)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@Time    :   2023/08/14 16:07:24
@Author  :   ChenHao
@Contact :   jerrychen1990@gmail.com
'''

import datetime
import time
import hashlib
from typing import List

import numpy as np
from snippets import dump_list, read2list


def save_csv_xls(df, path):
    return dump_list(df, path)


def read_csv_xls(path):
    return read2list(path)


def cal_vec_similarity(vec1: List[float], vec2: List[float], normalize=True, metric="cosine"):
    """
    计算两个向量之间的相似度
    :param vec1: 向量1
    :param vec2: 向量2
    :return: 相似度/距离
    """
    vec1 = np.array(vec1)
    vec2 = np.array(vec2)
    if normalize:
        vec1 = vec1 / np.linalg.norm(vec1, ord=2)
        vec2 = vec2 / np.linalg.norm(vec2, ord=2)
    if metric == "cosine":
        return np.dot(vec1, vec2) / (np.linalg.norm(vec1, 2) * np.linalg.norm(vec2, 2))
    if metric == "l2_distance":
        return np.linalg.norm((vec1-vec2), 2)
    else:
        raise Exception(f"Unknown metric: {metric}")


# 生成唯一request_id 根据时间戳+prompt+model生成
def gen_req_id(prompt, model):
    cur_time = datetime.datetime.fromtimestamp(time.time()).strftime('%Y%m%d-%H%M%S.%f')

    md5_hash = hashlib.md5(f"{prompt}={model}".encode('utf-8'))

    # 获取 MD5 摘要的字节串
    md5_digest = md5_hash.digest()
    md5_hex = md5_digest.hex()
    req_id = f"{md5_hex}_{cur_time}"
    return req_id



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@Time    :   2023/08/15 11:11:03
@Author  :   ChenHao
@Contact :   jerrychen1990@gmail.com
'''
import os
from loguru import logger
from snippets import print_info, set_logger

AGIT_ENV = os.environ.get("AGIT_ENV", "dev")
HOME = os.environ["HOME"]
AGIT_HOME = os.environ.get("AGIT_HOME", os.path.join(HOME, ".agit"))

AGIT_LOG_HOME = os.path.join(AGIT_HOME, "logs")
os.makedirs(AGIT_LOG_HOME, exist_ok=True)


set_logger(env=AGIT_ENV, module_name=__name__, log_dir=AGIT_LOG_HOME)


def show_env():
    print_info("current AGIT env", logger)
    logger.info(f"{AGIT_ENV=}")
    logger.info(f"{AGIT_HOME=}")
    logger.info(f"{AGIT_LOG_HOME=}")
    print_info("", logger)


show_env()



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@Time    :   2023/11/10 12:30:19
@Author  :   ChenHao
@Contact :   jerrychen1990@gmail.com
'''
import re
import requests
from loguru import logger


PREFIX_PROMPT_WITH_REF = "[Instruction]\nPlease act as an impartial judge and evaluate the quality of the response provided by an AI assistant to the user question displayed below. Your evaluation should consider factors such as the helpfulness, relevance, accuracy, depth, creativity, and level of detail of the response. Begin your evaluation by providing a short explanation. You will be given a high-quality reference answer and the assistant's answer. Be as objective as possible. You should first provide your explanation IN CHINESE, then you must rate the response on a scale of 1 to 10 by STRICTLY following the below MAPPING for the relation between the scores and response quality:\n1) The score 1~2 stands for very chaotic or absence of answer, and the AI assissant completely failed to address the instructions. Serious mathematical, logical and factual errors might also be included in this term. The gap between the AI assistant's answer and the high-quality reference answer is huge and insuperable.\n2) The score 3~4 indicates fragment-like responses from AI assistant's answer. It did not provide answers in proper grammar, fluency, or accuracy. There are obvious gaps between the high-quality reference answer and the AI assistant's response.\n3) The score 5~6 indicates for existence of minute disadvantage from the AI assistant's answer compared to the high-quality reference answer. Yet the AI assistant did provide an average answer. The AI assistant either did not fully satisfy instructions, or was somewhat short of helpfulness, relevance, depth, creativity, or detailedness. The disadvantages from the AI assistant's answer overwhelm its advantages.\n4) The score 7~8 indicates the AI assistant provided a good answer as well as the high-quality reference answer, satisfying the instruction, while addressing good helpfulness, relevance, accuracy, depth, creativity, and level of detail of the response. The AI assistant might have flaws compared to the reference answer, but that does not overwhelm the above advantages.\n5) The score 9~10 indicates the AI assistant responsed better than the provided reference answer in most aspects, fully achieved the instruction goal, and have unique advantages to the reference answer.\nYou should give scores around 7 if you do not find obvious advantages or disadvantages. You should seriously consider the above guide before give lowest and highest scores such as 1 or 10, and avoid such situation if you do not have sound explanations.\nAvoid any positional biases and ensure that the order in which the responses were presented does not influence your decision. Do not allow the length of the responses to influence your evaluation. Do not favor certain names of the assistants. AND again, VERY IMPORTANTLY, after you provide your explanation IN CHINESE, you must rate the response strictly following this format: \"[[rating]]\", for example: \"Rating: [[5]]\"."


def format_prompt_with_ref(question, reference, answer):
    return PREFIX_PROMPT_WITH_REF + f"\n\n[Question]\n{question}\n\n[The Start of Reference Answer]\n{reference}\n[The End of Reference Answer]\n\n[The Start of Assistant's Answer]\n{answer}\n[The End of Assistant's Answer]"


def eval_by_ref(question, reference, answer):
    logger.debug(f"Evaluating by reference answer: {reference}")
    prompt = format_prompt_with_ref(question, reference, answer)
    data = {
        "model": "chatglm2",
        "prompt": prompt,
        "seed": 34,
        "do_sample": False,
        "max_tokens": 256,
        "stream": False
    }
    header = {
        "Host": "api-research-score-32b-0926.glm.ai"
    }
    resp = requests.post(url="https://117.161.233.25:8443/v1/completions", verify=False, headers=header, json=data)
    resp.raise_for_status()
    text = resp.json()["choices"][0]["text"]
    # logger.info(text)
    idx = text.index("评分")
    reason = text[:idx].strip()
    pattern = "\[\[(.*?)\]\]"
    score = re.findall(pattern, text[idx + 3:])
    score = eval(score[0])

    # score = eval(text[idx + 3:])[0][0]
    return {"reason": reason, "score": score}


if __name__ == "__main__":
    print(eval_by_ref(question="中国人口有多少", reference="", answer="1亿"))

    # print(eval_by_ref(question="贵州茅台营收是多少", reference="100亿", answer="10000000000"))



# import agit.backend.openai_bk as openai_bk
# import agit.backend.zhipuai_bk as zhipuai_bk
import agit

def call_llm_api(prompt: str, model: str, **kwargs):
    if "chatglm" in model or "glm" in model: 
        if model == "chatglm3-6b":
            return agit.backend.openai_bk.call_llm_api(prompt=prompt, model=model, **kwargs)
        return agit.backend.zhipuai_bk.call_llm_api(prompt=prompt, model=model, **kwargs)
    else:
        return agit.backend.openai_bk.call_llm_api(prompt=prompt, model=model, **kwargs)
